----------------------------------------------------------------------------------
-- Engineer:       Stanislav Maslan, s.maslan@seznam.cz
-- www:            http://www.elektronika.kvalitne.cz
-- www:            http://www.elektronika.kvalitne.cz/PLA/TFT_tests/TFT_tests.html
--
-- Create Date:    15:08:21 05/01/2012 
-- Target Devices: CoolRunner2 XC2C256-7
-- Project Name:   640x480x16-bit TFT controller

-- Description:
-- ------------
-- Simple bitmap demo. CPLD is clocked at double rate than the TFT because I wasn't 
-- able to reach correct timing, so it requires 50MHz or so to reach recomended 
-- 25MHz pixel clock. 
----------------------------------------------------------------------------------
---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;


------ SELECT BITMAP ------
-- use work.bitmap_cr2.all;
use work.bitmap_cr2_big.all;
-- use work.bitmap_smile.all;

-- other libs
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity main is
	------ ports ------
	port(
		-- button input
		btn: in STD_LOGIC;
		-- source clock (50MHz)
		iclk: in STD_LOGIC;
		-- TFT pixel clock
		pclk: out STD_LOGIC;
		-- TFT synchronizations
		hsync: out STD_LOGIC;
		vsync: out STD_LOGIC;
		-- TFT color data
		red: out STD_LOGIC_VECTOR(4 downto 0);
		grn: out STD_LOGIC_VECTOR(5 downto 0);
		blu: out STD_LOGIC_VECTOR(4 downto 0));

	------ TFT definition constants ------
	-- TFT clocking parameters
	constant xw: integer := 790;
	constant yw: integer := 520;
	constant hsw: integer := 16;
	constant vsw: integer := 16;
	
	-- TFT visible area
	constant xofs: integer := 100;
	constant yofs: integer := 34;
	constant xsize: integer := 640;
	constant ysize: integer := 480;
	
	-- TFT color ranges
	constant rmax: integer := 31;
	constant gmax: integer := 63;
	constant bmax: integer := 31;
end main;


architecture Behavioral of main is
	-- internal clock
	signal xclk: bit;
	-- TFT position coutners
	signal pcnt: integer range 0 to xw;
	signal rcnt: integer range 0 to yw;
	-- color
	signal sred: integer range 0 to rmax := rmax;
	signal sgrn: integer range 0 to gmax := 0;
	signal sblu: integer range 0 to bmax := 0;
	-- bitmap position
	signal xpos: integer range 0 to xw := xofs + (xsize - bmpx)/2;
	signal ypos: integer range 0 to yw := yofs + (ysize - bmpy)/2;
	-- movement direction
	signal xdir: std_logic := '1';
	signal ydir: std_logic := '1';
	-- button latch
	signal pbtn: std_logic;
	-- background mode
	signal mode: std_logic := '0';
begin
	
	process(iclk,xclk,btn,pbtn,pcnt,rcnt,sred,sgrn,sblu,mode,xpos,ypos)
		variable col: integer range 0 to 2**bitx-1;
		variable row: integer range 0 to 2**bity-1;
		variable pix: std_logic;
		variable fsync: boolean;
	begin
		------ master clock divider ------
		if(iclk='0' and iclk'event) then
			-- internal clock signal
			xclk <= not xclk;
		end if;
			
		------ sync counters ------
		if(xclk='0' and xclk'event) then
			-- pixel counter
			if(pcnt>=xw) then
				pcnt <= 0;
				
				-- row counter
				if(rcnt>=yw) then
					rcnt <= 0;
				else
					rcnt <= rcnt + 1;
				end if;
			else
				pcnt <= pcnt + 1;
			end if;
		end if;
		
		------ internal frame sync signal
		fsync := pcnt=0 and rcnt=0;
				
		------ generate rainbow ------
		if(fsync and xclk='0' and xclk'event) then
			if(sred<rmax and sgrn=0 and sblu=bmax) then
				sred <= sred + 1;
			elsif(sred=rmax and sgrn<gmax and sblu=0) then
				sgrn <= sgrn + 1;
			elsif(sred=0 and sgrn=gmax and sblu<bmax) then
				sblu <= sblu + 1;
			elsif(sred>0 and sgrn=gmax and sblu=0) then
				sred <= sred - 1;
			elsif(sred=0 and sgrn>0 and sblu=bmax) then
				sgrn <= sgrn - 1;
			elsif(sred=rmax and sgrn=0 and sblu>0) then
				sblu <= sblu - 1;
			end if;
		end if;
		
		------ ping-pong movement ------
		if(fsync and xclk='0' and xclk'event) then
			-- x ping-pong
			if(xdir='1') then
				if(xpos<(xofs+xsize-bmpx)) then
					xpos <= xpos + 1;
				else
					xdir <= not xdir;
				end if;
			else
				if(xpos>xofs) then
					xpos <= xpos - 1;
				else
					xdir <= not xdir;
				end if;
			end if;
			
			-- y ping-pong
			if(ydir='1') then
				if(ypos<(yofs+ysize-bmpy)) then
					ypos <= ypos + 1;
				else
					ydir <= not ydir;
				end if;
			else
				if(ypos>yofs) then
					ypos <= ypos - 1;
				else
					ydir <= not ydir;
				end if;
			end if;			
		end if;

		------ bitmap position ------
		if(pcnt>=xpos and (pcnt-xpos)<bmpx) then
			col := pcnt - xpos;
		else
			col := bmpx;
		end if;
		if(rcnt>=ypos and (rcnt-ypos)<bmpy) then
			row := rcnt - ypos;
		else
			row := bmpy;
		end if;
		
		------ get bitmap pixel ------
		pix := pixels(conv_std_logic_vector(col,bitx),conv_std_logic_vector(row,bity));
				
		------ put it out using rainbow color ------
		if(pix='1') then
			-- active pixel
			red <= conv_std_logic_vector(sred,red'length);
			grn <= conv_std_logic_vector(sgrn,grn'length);
			blu <= conv_std_logic_vector(sblu,blu'length);			
		elsif(mode='0') then
			-- inactive black
			red <= conv_std_logic_vector(0,red'length);
			grn <= conv_std_logic_vector(0,grn'length);
			blu <= conv_std_logic_vector(0,blu'length);
		else
			-- inactive white
			red <= conv_std_logic_vector(rmax,red'length);
			grn <= conv_std_logic_vector(gmax,grn'length);
			blu <= conv_std_logic_vector(bmax,blu'length);
		end if;
		
		------ background mode switch ------
		-- latch button state
		if(fsync and xclk='0' and xclk'event) then
			pbtn <= btn;
		end if;
		
		-- change mode
		if(pbtn='0' and pbtn'event) then
			mode <= not mode;
		end if;
		
	end process;
	
	------ TFT clocks generator ------
	-- pixel clock
	pclk <= '1' when iclk='0' and xclk='1' else '0';
	
	-- hsync
	hsync <= '1' when pcnt<(xw-hsw) else '0';
			
	-- vsync
	vsync <= '1' when rcnt>vsw else '0';
		
end Behavioral;

